#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
# Imports
#

import sys
import os
from os.path import abspath, dirname

# Project must be built+installed to generate docs
import wpilib

# -- RTD configuration ------------------------------------------------

# on_rtd is whether we are on readthedocs.org, this line of code grabbed from docs.readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

# This is used for linking and such so we link to the thing we're building
rtd_version = os.environ.get("READTHEDOCS_VERSION", "latest")
if rtd_version not in ["stable", "latest"]:
    rtd_version = "stable"

# -- General configuration ------------------------------------------------


# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "robotpy_sphinx.all",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "RobotPy WPILib"
copyright = "2014-2020, RobotPy development team"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = ".".join(wpilib.__version__.split(".")[:2])
# The full version, including alpha/beta/rc tags.
release = wpilib.__version__

autoclass_content = "both"

intersphinx_mapping = {
    "networktables": (
        f"https://pynetworktables.readthedocs.io/en/{rtd_version}/",
        None,
    ),
    "wpimath": (
        f"https://robotpy.readthedocs.io/projects/wpimath/en/{rtd_version}/",
        None,
    ),
}

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output ----------------------------------------------

if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme

    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
else:
    html_theme = "default"

# Output file base name for HTML help builder.
htmlhelp_basename = "RobotPyWPILibdoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        "index",
        "RobotPyWPILib.tex",
        "RobotPy WPILib Documentation",
        "RobotPy development team",
        "manual",
    )
]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        "RobotPyWPILib",
        "RobotPy WPILib Documentation",
        "RobotPy development team",
        "RobotPyWPILib",
        "One line description of project.",
        "Miscellaneous",
    )
]

# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = "RobotPy WPILib"
epub_author = "RobotPy development team"
epub_publisher = "RobotPy development team"
epub_copyright = "2014, RobotPy development team"

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]

# -- Custom Document processing ----------------------------------------------

from robotpy_sphinx.regen import gen_package
from robotpy_sphinx.sidebar import generate_sidebar

generate_sidebar(
    globals(),
    "wpilib",
    "https://raw.githubusercontent.com/robotpy/docs-sidebar/master/sidebar.toml",
)

root = abspath(dirname(__file__))

gen_package(root, "wpilib", exclude=["wpi_*"])
gen_package(root, "wpilib.counter")
gen_package(root, "wpilib.event")
gen_package(root, "wpilib.drive")
gen_package(root, "wpilib.interfaces")
gen_package(root, "wpilib.simulation")



from . import _init_shuffleboard

# autogenerated by 'robotpy-build create-imports wpilib.shuffleboard'
from ._shuffleboard import (
    BuiltInLayouts,
    BuiltInWidgets,
    ComplexWidget,
    LayoutType,
    Shuffleboard,
    ShuffleboardComponentBase,
    ShuffleboardContainer,
    ShuffleboardEventImportance,
    ShuffleboardLayout,
    ShuffleboardTab,
    ShuffleboardValue,
    SimpleWidget,
    SuppliedBoolListValueWidget,
    SuppliedBoolValueWidget,
    SuppliedDoubleListValueWidget,
    SuppliedDoubleValueWidget,
    SuppliedFloatListValueWidget,
    SuppliedFloatValueWidget,
    SuppliedIntListValueWidget,
    SuppliedIntegerValueWidget,
    SuppliedRawValueWidget,
    SuppliedStringListValueWidget,
    SuppliedStringValueWidget,
    WidgetType,
    shuffleboardEventImportanceName,
)

__all__ = [
    "BuiltInLayouts",
    "BuiltInWidgets",
    "ComplexWidget",
    "LayoutType",
    "Shuffleboard",
    "ShuffleboardComponentBase",
    "ShuffleboardContainer",
    "ShuffleboardEventImportance",
    "ShuffleboardLayout",
    "ShuffleboardTab",
    "ShuffleboardValue",
    "SimpleWidget",
    "SuppliedBoolListValueWidget",
    "SuppliedBoolValueWidget",
    "SuppliedDoubleListValueWidget",
    "SuppliedDoubleValueWidget",
    "SuppliedFloatListValueWidget",
    "SuppliedFloatValueWidget",
    "SuppliedIntListValueWidget",
    "SuppliedIntegerValueWidget",
    "SuppliedRawValueWidget",
    "SuppliedStringListValueWidget",
    "SuppliedStringValueWidget",
    "WidgetType",
    "shuffleboardEventImportanceName",
]

del _init_shuffleboard



import pytest
import re
import weakref

import wpilib


def test_sendable_chooser():
    chooser = wpilib.SendableChooser()
    assert chooser.getSelected() is None

    chooser.setDefaultOption("option", True)
    assert chooser.getSelected() is True


def test_smart_dashboard_putdata():
    t = wpilib.Talon(4)
    ref = weakref.ref(t)
    wpilib.SmartDashboard.putData("talon", t)
    del t
    assert bool(ref) is True
    assert wpilib.SmartDashboard.getData("talon") is ref()


def test_motorcontrollergroup():
    t1 = wpilib.Talon(7)
    t2 = wpilib.Talon(8)
    g = wpilib.MotorControllerGroup(t1, t2)

    g.set(1)
    assert t1.get() == pytest.approx(1)
    assert t2.get() == pytest.approx(1)

    g.set(-1)
    assert t1.get() == pytest.approx(-1)
    assert t2.get() == pytest.approx(-1)


def test_motorcontrollergroup_error():
    with pytest.raises(
        TypeError, match=re.escape("Argument 1 must be a MotorController (got '1')")
    ):
        wpilib.MotorControllerGroup(1)



import hal
import wpilib
import logging
import threading
import time

from .report_error import reportError, reportErrorInternal


class Main:
    """
    Executes the robot code using the currently installed HAL (this is probably not what you want unless you're on the roboRIO)
    """

    def __init__(self, parser):
        pass

    def run(self, options, robot_class, **static_options):
        return robot_class.main(robot_class)


class RobotStarter:
    def __init__(self):
        self.logger = logging.getLogger("robotpy")
        self.robot = None
        self.suppressExitWarning = False

    def run(self, robot_cls: wpilib.RobotBase) -> bool:
        retval = False
        if hal.hasMain():
            rval = [False]

            def _start():
                try:
                    rval[0] = self.start(robot_cls)
                finally:
                    hal.exitMain()

            th = threading.Thread(target=_start, name="RobotThread", daemon=True)
            th.start()
            try:
                hal.runMain()
            except KeyboardInterrupt:
                self.logger.exception(
                    "THIS IS NOT AN ERROR: The user hit CTRL-C to kill the robot"
                )
                self.logger.info("Exiting because of keyboard interrupt")

            self.suppressExitWarning = True
            robot = self.robot
            if robot:
                try:
                    robot.endCompetition()
                except:
                    self.logger.warn("endCompetition raised an exception")

            th.join(1)
            if th.is_alive():
                self.logger.warn("robot thread didn't die, crash may occur next!")
            retval = rval[0]
        else:
            retval = self.start(robot_cls)

        from wpilib import RobotBase

        if RobotBase.isSimulation():
            import wpilib.simulation

            wpilib.simulation._simulation._resetMotorSafety()

        return retval

    def start(self, robot_cls: wpilib.RobotBase) -> bool:
        try:
            return self._start(robot_cls)
        except:
            reportErrorInternal(
                "The robot program quit unexpectedly. This is usually due to a code error.\n"
                "The above stacktrace can help determine where the error occurred.\n",
                True,
            )
            return False

    def _start(self, robot_cls: wpilib.RobotBase) -> bool:

        hal.report(
            hal.tResourceType.kResourceType_Language,
            hal.tInstances.kLanguage_Python,
            0,
            wpilib.__version__,
        )

        if not wpilib.Notifier.setHALThreadPriority(True, 40):
            reportErrorInternal(
                "Setting HAL Notifier RT priority to 40 failed", isWarning=True
            )

        isSimulation = wpilib.RobotBase.isSimulation()

        # hack: initialize networktables before creating the robot
        #       class, otherwise our logger doesn't get created
        import ntcore

        inst = ntcore.NetworkTableInstance.getDefault()

        # subscribe to "" to force persistent values to progagate to local
        msub = ntcore.MultiSubscriber(inst, [""])

        if not isSimulation:
            inst.startServer("/home/lvuser/networktables.ini")
        else:
            inst.startServer()

        # wait for the NT server to actually start
        for i in range(100):
            if (
                inst.getNetworkMode()
                & ntcore.NetworkTableInstance.NetworkMode.kNetModeStarting
            ) == 0:
                break
            # real sleep since we're waiting for the server, not simulated sleep
            time.sleep(0.010)
        else:
            reportErrorInternal(
                "timed out while waiting for NT server to start", isWarning=True
            )

        wpilib.SmartDashboard.init()

        # Call DriverStation.refreshData() to kick things off
        wpilib.DriverStation.refreshData()

        try:
            self.robot = robot_cls()
        except:
            reportError(
                f"Unhandled exception instantiating robot {robot_cls.__name__}", True
            )
            reportErrorInternal(f"Could not instantiate robot {robot_cls.__name__}!")
            raise

        # TODO: Add a check to see if the user forgot to call super().__init__()
        # if not hasattr(robot, "_RobotBase__initialized"):
        #     logger.error(
        #         "If your robot class has an __init__ function, it must call super().__init__()!"
        #     )
        #     return False

        if not isSimulation:
            try:
                from robotpy.version import __version__ as robotpy_version

                version_string = f"RobotPy {robotpy_version}"
            except ImportError:
                version_string = f"robotpy-wpilib {wpilib.__version__}"

            try:
                with open("/tmp/frc_versions/FRC_Lib_Version.ini", "w") as fp:
                    fp.write(version_string)
            except:
                reportErrorInternal("Could not write FRC version file to disk")

        try:
            self.robot.startCompetition()
        except KeyboardInterrupt:
            self.robot = None
            self.logger.exception(
                "THIS IS NOT AN ERROR: The user hit CTRL-C to kill the robot"
            )
            self.logger.info("Exiting because of keyboard interrupt")
            return True
        except:
            self.robot = None

            reportError("Unhandled exception", True)
            raise
        else:
            self.robot = None
            if self.suppressExitWarning:
                self.logger.info("Robot code exited")
                return True
            else:
                # startCompetition never returns unless exception occurs....
                reportError("Unexpected return from startCompetition() method.", False)
                return False



import hal
import sys
import traceback
import logging
import inspect

robotpy_logger = logging.getLogger("robotpy")
user_logger = logging.getLogger("your.robot")


def reportErrorInternal(
    error: str,
    printTrace: bool = False,
    fromUser: bool = False,
    isWarning: bool = True,
    code: int = 1,
) -> None:
    traceString = ""

    if fromUser:
        log = user_logger
    else:
        log = robotpy_logger

    if printTrace:
        exc_info = sys.exc_info()

        exc = exc_info[0]
        if exc is None:
            tb = traceback.extract_stack()[:-2]
        else:
            tb = traceback.extract_tb(exc_info[2])

        locString = "%s.%s:%s" % (tb[-1][0], tb[-1][1], tb[-1][2])

        trc = "Traceback (most recent call last):\n"
        stackstr = trc + "".join(traceback.format_list(tb))
        if exc is not None:
            stackstr += "  " + ("".join(traceback.format_exception(*exc_info))).lstrip(
                trc
            )
        traceString += "\n" + stackstr

        if exc is None:
            log.error(error + "\n" + traceString)
        else:
            log.error(error, exc_info=exc_info)
    else:
        if isWarning:
            log.warning(error)
        else:
            log.error(error)

        try:
            frame = inspect.stack(context=2)[-1]
            locString = f"{frame.filename}:{frame.lineno}"
        except Exception:
            locString = "<unknown location>"

    if not hal.__hal_simulation__:
        hal.sendError(
            not isWarning,
            code,
            False,
            error.encode("utf-8"),
            locString.encode("utf-8"),
            traceString.encode("utf-8"),
            True,
        )


def reportError(error: str, printTrace: bool = False) -> None:
    """
    Report error to Driver Station, and also prints error to ``sys.stderr``.
    Optionally appends stack trace to error message.

    :param error: message to show
    :param printTrace: If True, appends stack trace to error string
    """
    reportErrorInternal(error, printTrace, fromUser=True)


def reportWarning(error: str, printTrace: bool = False) -> None:
    """
    Report warning to Driver Station, and also prints error to ``sys.stderr``.
    Optionally appends stack trace to error message.

    :param error: message to show
    :param printTrace: If True, appends stack trace to error string
    """
    reportErrorInternal(error, printTrace, fromUser=True, isWarning=True)



import pathlib
import wpilib
import pytest
import sys


@pytest.mark.skipif(sys.platform == "darwin", reason="DataLogManager crashes on exit")
def test_get_log(tmp_path: pathlib.Path):
    log_dir = tmp_path / "wpilogs"
    log_dir.mkdir()
    wpilib.DataLogManager.start(str(log_dir))
    log = wpilib.DataLogManager.getLog()
    assert log is not None



from typing import Optional

import hal
import subprocess
import threading

import logging

logger = logging.getLogger("wpilib.cs")

__all__ = ["CameraServer"]


class CameraServer:
    """
    Provides a way to launch an out of process cscore-based camera
    service instance, for streaming or for image processing.

    .. note:: This does not correspond directly to the wpilib
              CameraServer object; that can be found as
              :class:`cscore.CameraServer`. However, you should
              not use cscore directly from your robot code, see
              the documentation for details
    """

    _alive = False
    _launched = False

    @classmethod
    def is_alive(cls) -> bool:
        """:returns: True if the CameraServer is still alive"""
        return cls._alive

    @classmethod
    def launch(cls, vision_py: Optional[str] = None) -> None:
        """
        Launches the CameraServer process in autocapture mode or
        using a user-specified python script

        :param vision_py: If specified, this is the relative path to
                          a filename with a function in it

        Example usage::

            wpilib.CameraServer.launch("vision.py:main")

        .. warning:: You must have robotpy-cscore installed, or this
                     function will fail without returning an error
                     (you will see an error in the console).

        """

        if cls._launched:
            return

        cls._launched = True
        from ._wpilib import RobotBase

        if RobotBase.isSimulation():
            logger.info("Would launch CameraServer with vision_py=%s", vision_py)
            cls._alive = True
        else:
            logger.info("Launching CameraServer process")

            # Launch the cscore launcher in a separate process
            import sys

            args = [sys.executable, "-m", "cscore"]

            # TODO: Get accurate reporting data from the other cscore process. For
            # now, just differentiate between users with a custom py file and those
            # who do not. cscore handle values indicate type with bits 24-30

            if vision_py:
                if not vision_py.startswith("/"):
                    vision_py = "/home/lvuser/py/" + vision_py
                args.append(vision_py)
                hal.report(hal.tResourceType.kResourceType_PCVideoServer, 0x51)
            else:
                hal.report(hal.tResourceType.kResourceType_PCVideoServer, 0x52)

            # We open a pipe to it so that when this process exits, it dies
            proc = subprocess.Popen(
                args, close_fds=True, stdin=subprocess.PIPE, cwd="/home/lvuser/py"
            )
            th = threading.Thread(target=cls._monitor_child, args=(proc,))
            th.daemon = True
            th.start()

    @classmethod
    def _monitor_child(cls, proc: subprocess.Popen) -> None:
        proc.wait()
        logger.warning("CameraServer process exited with exitcode %s", proc.returncode)
        cls._alive = False



from . import _init_counter

# autogenerated by 'robotpy-build create-imports wpilib.counter'
from ._counter import (
    EdgeConfiguration,
    ExternalDirectionCounter,
    Tachometer,
    UpDownCounter,
)

__all__ = [
    "EdgeConfiguration",
    "ExternalDirectionCounter",
    "Tachometer",
    "UpDownCounter",
]
